// Automaticly generaged by C:\Work\Projects\Delegates\include\srutil\delegate\detail\make_delegates.pl

// 0 params
#define VIGRA_DELEGATE_PARAM_COUNT 0
#define VIGRA_DELEGATE_TEMPLATE_PARAMS
#define VIGRA_DELEGATE_TEMPLATE_ARGS
#define VIGRA_DELEGATE_PARAMS
#define VIGRA_DELEGATE_ARGS
#define VIGRA_DELEGATE_INVOKER_INITIALIZATION_LIST
#define VIGRA_DELEGATE_INVOKER_DATA
#include "delegate_template.hxx"
#undef VIGRA_DELEGATE_PARAM_COUNT
#undef VIGRA_DELEGATE_TEMPLATE_PARAMS
#undef VIGRA_DELEGATE_TEMPLATE_ARGS
#undef VIGRA_DELEGATE_PARAMS
#undef VIGRA_DELEGATE_ARGS
#undef VIGRA_DELEGATE_INVOKER_INITIALIZATION_LIST
#undef VIGRA_DELEGATE_INVOKER_DATA

// 1 params
#define VIGRA_DELEGATE_PARAM_COUNT 1
#define VIGRA_DELEGATE_TEMPLATE_PARAMS typename A1
#define VIGRA_DELEGATE_TEMPLATE_ARGS A1
#define VIGRA_DELEGATE_PARAMS A1 a1
#define VIGRA_DELEGATE_ARGS a1
#define VIGRA_DELEGATE_INVOKER_INITIALIZATION_LIST a1(a1)
#define VIGRA_DELEGATE_INVOKER_DATA A1 a1;
#include "delegate_template.hxx"
#undef VIGRA_DELEGATE_PARAM_COUNT
#undef VIGRA_DELEGATE_TEMPLATE_PARAMS
#undef VIGRA_DELEGATE_TEMPLATE_ARGS
#undef VIGRA_DELEGATE_PARAMS
#undef VIGRA_DELEGATE_ARGS
#undef VIGRA_DELEGATE_INVOKER_INITIALIZATION_LIST
#undef VIGRA_DELEGATE_INVOKER_DATA

// 2 params
#define VIGRA_DELEGATE_PARAM_COUNT 2
#define VIGRA_DELEGATE_TEMPLATE_PARAMS typename A1, typename A2
#define VIGRA_DELEGATE_TEMPLATE_ARGS A1, A2
#define VIGRA_DELEGATE_PARAMS A1 a1, A2 a2
#define VIGRA_DELEGATE_ARGS a1,a2
#define VIGRA_DELEGATE_INVOKER_INITIALIZATION_LIST a1(a1),a2(a2)
#define VIGRA_DELEGATE_INVOKER_DATA A1 a1;A2 a2;
#include "delegate_template.hxx"
#undef VIGRA_DELEGATE_PARAM_COUNT
#undef VIGRA_DELEGATE_TEMPLATE_PARAMS
#undef VIGRA_DELEGATE_TEMPLATE_ARGS
#undef VIGRA_DELEGATE_PARAMS
#undef VIGRA_DELEGATE_ARGS
#undef VIGRA_DELEGATE_INVOKER_INITIALIZATION_LIST
#undef VIGRA_DELEGATE_INVOKER_DATA

// 3 params
#define VIGRA_DELEGATE_PARAM_COUNT 3
#define VIGRA_DELEGATE_TEMPLATE_PARAMS typename A1, typename A2, typename A3
#define VIGRA_DELEGATE_TEMPLATE_ARGS A1, A2, A3
#define VIGRA_DELEGATE_PARAMS A1 a1, A2 a2, A3 a3
#define VIGRA_DELEGATE_ARGS a1,a2,a3
#define VIGRA_DELEGATE_INVOKER_INITIALIZATION_LIST a1(a1),a2(a2),a3(a3)
#define VIGRA_DELEGATE_INVOKER_DATA A1 a1;A2 a2;A3 a3;
#include "delegate_template.hxx"
#undef VIGRA_DELEGATE_PARAM_COUNT
#undef VIGRA_DELEGATE_TEMPLATE_PARAMS
#undef VIGRA_DELEGATE_TEMPLATE_ARGS
#undef VIGRA_DELEGATE_PARAMS
#undef VIGRA_DELEGATE_ARGS
#undef VIGRA_DELEGATE_INVOKER_INITIALIZATION_LIST
#undef VIGRA_DELEGATE_INVOKER_DATA

// 4 params
#define VIGRA_DELEGATE_PARAM_COUNT 4
#define VIGRA_DELEGATE_TEMPLATE_PARAMS typename A1, typename A2, typename A3, typename A4
#define VIGRA_DELEGATE_TEMPLATE_ARGS A1, A2, A3, A4
#define VIGRA_DELEGATE_PARAMS A1 a1, A2 a2, A3 a3, A4 a4
#define VIGRA_DELEGATE_ARGS a1,a2,a3,a4
#define VIGRA_DELEGATE_INVOKER_INITIALIZATION_LIST a1(a1),a2(a2),a3(a3),a4(a4)
#define VIGRA_DELEGATE_INVOKER_DATA A1 a1;A2 a2;A3 a3;A4 a4;
#include "delegate_template.hxx"
#undef VIGRA_DELEGATE_PARAM_COUNT
#undef VIGRA_DELEGATE_TEMPLATE_PARAMS
#undef VIGRA_DELEGATE_TEMPLATE_ARGS
#undef VIGRA_DELEGATE_PARAMS
#undef VIGRA_DELEGATE_ARGS
#undef VIGRA_DELEGATE_INVOKER_INITIALIZATION_LIST
#undef VIGRA_DELEGATE_INVOKER_DATA

// 5 params
#define VIGRA_DELEGATE_PARAM_COUNT 5
#define VIGRA_DELEGATE_TEMPLATE_PARAMS typename A1, typename A2, typename A3, typename A4, typename A5
#define VIGRA_DELEGATE_TEMPLATE_ARGS A1, A2, A3, A4, A5
#define VIGRA_DELEGATE_PARAMS A1 a1, A2 a2, A3 a3, A4 a4, A5 a5
#define VIGRA_DELEGATE_ARGS a1,a2,a3,a4,a5
#define VIGRA_DELEGATE_INVOKER_INITIALIZATION_LIST a1(a1),a2(a2),a3(a3),a4(a4),a5(a5)
#define VIGRA_DELEGATE_INVOKER_DATA A1 a1;A2 a2;A3 a3;A4 a4;A5 a5;
#include "delegate_template.hxx"
#undef VIGRA_DELEGATE_PARAM_COUNT
#undef VIGRA_DELEGATE_TEMPLATE_PARAMS
#undef VIGRA_DELEGATE_TEMPLATE_ARGS
#undef VIGRA_DELEGATE_PARAMS
#undef VIGRA_DELEGATE_ARGS
#undef VIGRA_DELEGATE_INVOKER_INITIALIZATION_LIST
#undef VIGRA_DELEGATE_INVOKER_DATA

// 6 params
#define VIGRA_DELEGATE_PARAM_COUNT 6
#define VIGRA_DELEGATE_TEMPLATE_PARAMS typename A1, typename A2, typename A3, typename A4, typename A5, typename A6
#define VIGRA_DELEGATE_TEMPLATE_ARGS A1, A2, A3, A4, A5, A6
#define VIGRA_DELEGATE_PARAMS A1 a1, A2 a2, A3 a3, A4 a4, A5 a5, A6 a6
#define VIGRA_DELEGATE_ARGS a1,a2,a3,a4,a5,a6
#define VIGRA_DELEGATE_INVOKER_INITIALIZATION_LIST a1(a1),a2(a2),a3(a3),a4(a4),a5(a5),a6(a6)
#define VIGRA_DELEGATE_INVOKER_DATA A1 a1;A2 a2;A3 a3;A4 a4;A5 a5;A6 a6;
#include "delegate_template.hxx"
#undef VIGRA_DELEGATE_PARAM_COUNT
#undef VIGRA_DELEGATE_TEMPLATE_PARAMS
#undef VIGRA_DELEGATE_TEMPLATE_ARGS
#undef VIGRA_DELEGATE_PARAMS
#undef VIGRA_DELEGATE_ARGS
#undef VIGRA_DELEGATE_INVOKER_INITIALIZATION_LIST
#undef VIGRA_DELEGATE_INVOKER_DATA

// 7 params
#define VIGRA_DELEGATE_PARAM_COUNT 7
#define VIGRA_DELEGATE_TEMPLATE_PARAMS typename A1, typename A2, typename A3, typename A4, typename A5, typename A6, typename A7
#define VIGRA_DELEGATE_TEMPLATE_ARGS A1, A2, A3, A4, A5, A6, A7
#define VIGRA_DELEGATE_PARAMS A1 a1, A2 a2, A3 a3, A4 a4, A5 a5, A6 a6, A7 a7
#define VIGRA_DELEGATE_ARGS a1,a2,a3,a4,a5,a6,a7
#define VIGRA_DELEGATE_INVOKER_INITIALIZATION_LIST a1(a1),a2(a2),a3(a3),a4(a4),a5(a5),a6(a6),a7(a7)
#define VIGRA_DELEGATE_INVOKER_DATA A1 a1;A2 a2;A3 a3;A4 a4;A5 a5;A6 a6;A7 a7;
#include "delegate_template.hxx"
#undef VIGRA_DELEGATE_PARAM_COUNT
#undef VIGRA_DELEGATE_TEMPLATE_PARAMS
#undef VIGRA_DELEGATE_TEMPLATE_ARGS
#undef VIGRA_DELEGATE_PARAMS
#undef VIGRA_DELEGATE_ARGS
#undef VIGRA_DELEGATE_INVOKER_INITIALIZATION_LIST
#undef VIGRA_DELEGATE_INVOKER_DATA

// 8 params
#define VIGRA_DELEGATE_PARAM_COUNT 8
#define VIGRA_DELEGATE_TEMPLATE_PARAMS typename A1, typename A2, typename A3, typename A4, typename A5, typename A6, typename A7, typename A8
#define VIGRA_DELEGATE_TEMPLATE_ARGS A1, A2, A3, A4, A5, A6, A7, A8
#define VIGRA_DELEGATE_PARAMS A1 a1, A2 a2, A3 a3, A4 a4, A5 a5, A6 a6, A7 a7, A8 a8
#define VIGRA_DELEGATE_ARGS a1,a2,a3,a4,a5,a6,a7,a8
#define VIGRA_DELEGATE_INVOKER_INITIALIZATION_LIST a1(a1),a2(a2),a3(a3),a4(a4),a5(a5),a6(a6),a7(a7),a8(a8)
#define VIGRA_DELEGATE_INVOKER_DATA A1 a1;A2 a2;A3 a3;A4 a4;A5 a5;A6 a6;A7 a7;A8 a8;
#include "delegate_template.hxx"
#undef VIGRA_DELEGATE_PARAM_COUNT
#undef VIGRA_DELEGATE_TEMPLATE_PARAMS
#undef VIGRA_DELEGATE_TEMPLATE_ARGS
#undef VIGRA_DELEGATE_PARAMS
#undef VIGRA_DELEGATE_ARGS
#undef VIGRA_DELEGATE_INVOKER_INITIALIZATION_LIST
#undef VIGRA_DELEGATE_INVOKER_DATA

// 9 params
#define VIGRA_DELEGATE_PARAM_COUNT 9
#define VIGRA_DELEGATE_TEMPLATE_PARAMS typename A1, typename A2, typename A3, typename A4, typename A5, typename A6, typename A7, typename A8, typename A9
#define VIGRA_DELEGATE_TEMPLATE_ARGS A1, A2, A3, A4, A5, A6, A7, A8, A9
#define VIGRA_DELEGATE_PARAMS A1 a1, A2 a2, A3 a3, A4 a4, A5 a5, A6 a6, A7 a7, A8 a8, A9 a9
#define VIGRA_DELEGATE_ARGS a1,a2,a3,a4,a5,a6,a7,a8,a9
#define VIGRA_DELEGATE_INVOKER_INITIALIZATION_LIST a1(a1),a2(a2),a3(a3),a4(a4),a5(a5),a6(a6),a7(a7),a8(a8),a9(a9)
#define VIGRA_DELEGATE_INVOKER_DATA A1 a1;A2 a2;A3 a3;A4 a4;A5 a5;A6 a6;A7 a7;A8 a8;A9 a9;
#include "delegate_template.hxx"
#undef VIGRA_DELEGATE_PARAM_COUNT
#undef VIGRA_DELEGATE_TEMPLATE_PARAMS
#undef VIGRA_DELEGATE_TEMPLATE_ARGS
#undef VIGRA_DELEGATE_PARAMS
#undef VIGRA_DELEGATE_ARGS
#undef VIGRA_DELEGATE_INVOKER_INITIALIZATION_LIST
#undef VIGRA_DELEGATE_INVOKER_DATA

// 10 params
#define VIGRA_DELEGATE_PARAM_COUNT 10
#define VIGRA_DELEGATE_TEMPLATE_PARAMS typename A1, typename A2, typename A3, typename A4, typename A5, typename A6, typename A7, typename A8, typename A9, typename A10
#define VIGRA_DELEGATE_TEMPLATE_ARGS A1, A2, A3, A4, A5, A6, A7, A8, A9, A10
#define VIGRA_DELEGATE_PARAMS A1 a1, A2 a2, A3 a3, A4 a4, A5 a5, A6 a6, A7 a7, A8 a8, A9 a9, A10 a10
#define VIGRA_DELEGATE_ARGS a1,a2,a3,a4,a5,a6,a7,a8,a9,a10
#define VIGRA_DELEGATE_INVOKER_INITIALIZATION_LIST a1(a1),a2(a2),a3(a3),a4(a4),a5(a5),a6(a6),a7(a7),a8(a8),a9(a9),a10(a10)
#define VIGRA_DELEGATE_INVOKER_DATA A1 a1;A2 a2;A3 a3;A4 a4;A5 a5;A6 a6;A7 a7;A8 a8;A9 a9;A10 a10;
#include "delegate_template.hxx"
#undef VIGRA_DELEGATE_PARAM_COUNT
#undef VIGRA_DELEGATE_TEMPLATE_PARAMS
#undef VIGRA_DELEGATE_TEMPLATE_ARGS
#undef VIGRA_DELEGATE_PARAMS
#undef VIGRA_DELEGATE_ARGS
#undef VIGRA_DELEGATE_INVOKER_INITIALIZATION_LIST
#undef VIGRA_DELEGATE_INVOKER_DATA

